unit DataObject;

interface

uses Windows, SysUtils, ComObj, ActiveX, mmc;

const
//
// This is the minimum set of clipboard formats we must implement.
// MMC uses these to get necessary information from our snapin about
// our nodes.
//
  s_cfDisplayName    : Integer = 0;
  s_cfNodeType       : Integer = 0;
  s_cfCoClass        : Integer = 0;
  s_cfInternal       : Integer = 0;
  s_cfMultiSel       : Integer = 0;
  s_cfNodeTypeString : Integer = 0;
  s_cfWorkstation    : Integer = 0;

const
  cNodeTypeStatic     : TGuid = '{44092d22-1d7e-11d0-a29b-00c04fd909dd}';
  csNodeTypeStatic    : PWideChar = '{44092d22-1d7e-11d0-a29b-00c04fd909dd}';
  cNodeTypeCompany    : TGuid = '{55092d22-1d7e-11d0-a29b-00c04fd909dd}';
  csNodeTypeCompany   : PWideChar = '{55092d22-1d7e-11d0-a29b-00c04fd909dd}';
  cNodeTypeUser       : TGuid = '{66092d22-1d7e-11d0-a29b-00c04fd909dd}';
  csNodeTypeUser      : PWideChar = '{66092d22-1d7e-11d0-a29b-00c04fd909dd}';
  cNodeTypeExtCompany : TGuid = '{e5092d22-1d7e-11d0-a29b-00c04fd909dd}';
  csNodeTypeExtCompany: PWideChar = '{e5092d22-1d7e-11d0-a29b-00c04fd909dd}';
  cNodeTypeExtUser    : TGuid = '{e6092d22-1d7e-11d0-a29b-00c04fd909dd}';
  csNodeTypeExtUser   : PWideChar = '{e6092d22-1d7e-11d0-a29b-00c04fd909dd}';
  cNodeTypeVirtual    : TGuid = '{AD003E5A-0483-11d1-AEF7-00C04FB6DD2C}';
  csNodeTypeVirtual   : PWideChar = '{AD003E5A-0483-11d1-AEF7-00C04FB6DD2C}';
  cNodeTypeDynamic    : TGuid = '{0ac69b7a-afce-11d0-a79b-00c04fd8d565}';
  csNodeTypeDynamic   : PWideChar = '{0ac69b7a-afce-11d0-a79b-00c04fd8d565}';

type
  TCookieType = (COOKIE_IS_ROOT, COOKIE_IS_STATUS);

  TInternal = object
    m_type : DATA_OBJECT_TYPES;         // What context is the data object.
    m_cookie : LongInt;                 // What object the cookie represents
    m_string : string;                  //
    m_clsid : TGuid;                    // Class ID of who created this data object

    constructor Create;
    procedure Assign (const Another : TInternal);
    function Equals (const Another : TInternal) : boolean;
  end;
  PInternal = ^TInternal;

  TDataObject = class(TInterfacedObject, IDataObject)
  private
    fInternal : TInternal;

    fpMultiSelData : PBYTE;
    fcbMultiSelData : Integer;
    fbMultiSelDobj : BOOL;

    function GetData(const formatetcIn: TFormatEtc; out medium: TStgMedium): HResult; stdcall;
    function EnumFormatEtc (dwDirection: Longint; out enumFormatEtc: IEnumFormatEtc): HResult; stdcall;
    function QueryGetData(const formatetc: TFormatEtc): HResult; stdcall;
    function GetCanonicalFormatEtc(const formatetc: TFormatEtc; out formatetcOut: TFormatEtc): HResult; stdcall;
    function SetData(const formatetc: TFormatEtc; var medium: TStgMedium; fRelease: BOOL): HResult; stdcall;
    function DAdvise(const formatetc: TFormatEtc; advf: Longint; const advSink: IAdviseSink; out dwConnection: Longint): HResult; stdcall;
    function DUnadvise(dwConnection: Longint): HResult; stdcall;
    function EnumDAdvise(out enumAdvise: IEnumStatData): HResult; stdcall;
    function CreateData               (buffer : pointer; len : Integer; var medium : TstgMedium) : HRESULT;
    function CreateNodeTypeData       (var medium : TstgMedium) : HRESULT;
    function CreateCoClassID          (var medium : TstgMedium) : HRESULT;
    function CreateNodeTypeStringData (var medium : TstgMedium) : HRESULT;
    function CreateDisplayName        (var medium : TstgMedium) : HRESULT;
    function CreateInternal           (var medium : TstgMedium) : HRESULT;
    function CreateWorkstationName    (var medium : TstgMedium) : HRESULT;

    procedure SetCookie (value : LongInt);
    function GetCookie : LongInt;
    procedure SetString (const value : string);
    function GetString : string;
    procedure SetCLSID (const value : TGuid);
    function GetCLSID : TGuid;
    procedure SetObjectType (value : DATA_OBJECT_TYPES);
    function GetObjectType : DATA_OBJECT_TYPES;
  public


    constructor Create;
    destructor Destroy; override;

    property Cookie : LongInt read GetCookie write SetCookie;
    property Strng : string read GetString write SetString;
    property CLSID : TGuid read GetCLSID write SetCLSID;
    property ObjectType : DATA_OBJECT_TYPES read GetObjectType write SetObjectType;

    procedure SetMultiSelData (pbMultiSelData : PBYTE; cbMultiSelData : Integer);
    procedure SetMultiSelDobj;


    //
    // IDataObject overrides
    //

    function GetDataHere (const formatetc: TFormatEtc; out medium: TStgMedium) : HRESULT; stdcall;


private
end;
PDataObject = ^TDataObject;
PPDataObject = ^PDataObject;

function ExtractOwnDataObject (dataObject : IDataObject)  : TDataObject;
function ExtractInternalFormat (dataObject : IDataObject) : PInternal;
function ExtractClassId (dataObject : IDataObject) : PGuid;
function ExtractNodeType (dataObject : IDataObject) : PGuid;
function ExtractWorkstation (dataObject : IDataObject) : PWideChar;

implementation

uses service;

// Generated with uuidgen. Each node must have a GUID associated with it.
// This one is for the main root node.

resourcestring
  idsNodeName = 'Test Node';

const
  cObjectTypeResultItem : TGuid = '{00c86e52-af90-11d0-a79b-00c04fd8d565}';
  csObjectTypeResultItem: PWideChar = '{00c86e52-af90-11d0-a79b-00c04fd8d565}';

  CF_SNAPIN_INTERNAL = 'SNAPIN_INTERNAL';
  CF_SNAPIN_WORKSTATION = 'SNAPIN_WORKSTATION';

constructor TInternal.Create;
begin
  m_type := CCT_UNINITIALIZED;
  m_cookie := -1;
  FillChar (m_clsid, sizeof (m_clsid), 0);
end;

procedure TInternal.Assign (const Another : TInternal);
begin
  m_type := Another.m_type;
  m_cookie := Another.m_cookie;
  m_string := Another.m_string;
  Move (Another.m_clsid, m_clsid, sizeof (m_clsid))
end;

function TInternal.Equals (const Another : TInternal) : boolean;
begin
 result := m_string = Another.m_string
end;

constructor TDataObject.Create;
begin
  inherited Create;
  fInternal.Create;

  fpMultiSelData := Nil;
  fcbMultiSelData := 0;
  fbMultiSelDobj := FALSE;

  s_cfInternal    := RegisterClipboardFormat (CF_SNAPIN_INTERNAL);
  s_cfDisplayName := RegisterClipboardFormat (CCF_DISPLAY_NAME);
  s_cfNodeType    := RegisterClipboardFormat (CCF_NODETYPE);
  s_cfCoClass     := RegisterClipboardFormat (CCF_SNAPIN_CLASSID);
  s_cfMultiSel    := RegisterClipboardFormat (CCF_OBJECT_TYPES_IN_MULTI_SELECT);
  s_cfWorkstation := RegisterClipboardFormat (CF_SNAPIN_WORKSTATION);
end;


//+--------------------------------------------------------------------------
//
//  Member:     CDataObject::~CDataObject
//
//  Synopsis:   dtor
//
//---------------------------------------------------------------------------

destructor TDataObject.Destroy;
begin
  if Assigned (fpMultiSelData) then
    FreeMem (fpMultiSelData);
  inherited
end;


function TDataObject.GetDataHere (const formatetc: TFormatEtc; out medium: TStgMedium) : HRESULT;
var
  cf : word;
begin
  result := DV_E_CLIPFORMAT;
  cf := formatetc.cfFormat;

  if cf = s_cfNodeType then
    result := CreateNodeTypeData (medium)
  else
    if cf = s_cfCoClass then
      result := CreateCoClassID (medium)
    else
      if cf = s_cfNodeTypeString then
        result := CreateNodeTypeStringData (medium)
      else
        if cf = s_cfDisplayName then
          result := CreateDisplayName (medium)
        else
          if cf = s_cfInternal then
            result := CreateInternal (medium)
          else
            if cf = s_cfWorkstation then
              result := CreateWorkstationName (medium)
end;

function TDataObject.CreateData (buffer : pointer; len : Integer; var medium : TStgMedium) : HRESULT;
var
  stream : IStream;
  written : LongInt;
begin

  if (not Assigned (buffer)) then
     result := E_POINTER
  else
  begin
    result := DV_E_TYMED;

    if medium.tymed = TYMED_HGLOBAL then
    begin
      result := CreateStreamOnHGlobal (medium.hGlobal, FALSE, stream);

      if result = S_OK then
      begin
        result := stream.Write (buffer, len, @written);
      end
    end
  end
end;

function TDataObject.CreateNodeTypeData (var medium : TStgMedium) : HRESULT;
var
  pcObjectType : PGuid;
  folder : TFolder;
begin
  result := S_OK;
  pcObjectType := Nil;
  if fInternal.m_cookie = 0 then
    pcObjectType := @cNodeTypeStatic
  else
    if fInternal.m_type = CCT_SCOPE then
    begin
      folder := TFolder (fInternal.m_cookie);
      Assert (Assigned (folder));
      if not Assigned (folder) then
        result := E_UNEXPECTED
      else
        case folder.FolderType of
          ftCOMPANY     : pcObjectType := @cNodeTypeCompany;
          ftUSER        : pcObjectType := @cNodeTypeUser;
          ftEXT_COMPANY : pcObjectType := @cNodeTypeExtCompany;
          ftEXT_USER    : pcObjectType := @cNodeTypeExtUser;
          ftVIRTUAL,
          ftEXT_VIRTUAL : pcObjectType := @cNodeTypeVirtual;
          else            pcObjectType := @cNodeTypeDynamic
        end
    end
    else
      if fInternal.m_type = CCT_RESULT then
      begin
        pcObjectType := @cObjectTypeResultItem
      end;

  if result = S_OK then
    result := CreateData (pcObjectType, sizeof (TGuid), medium);
end;

function TDataObject.CreateCoClassID (var medium : TStgMedium) : HRESULT;
begin
  result :=CreateData (@fInternal.m_clsid, sizeof (TGuid), medium)
end;

function TDataObject.CreateNodeTypeStringData (var medium : TStgMedium) : HRESULT;
var
  cObjectType : PWideChar;
  folder : TFolder;
begin
  cObjectType := Nil;
  result := S_OK;

  if fInternal.m_cookie = 0 then
    cObjectType := csNodeTypeStatic
  else
    if fInternal.m_type = CCT_SCOPE then
    begin
      folder := TFolder (fInternal.m_cookie);

      Assert (Assigned (folder));
      if not Assigned (folder) then
        result := E_UNEXPECTED
      else
      begin
        case folder.FolderType of
          ftCOMPANY     : cObjectType := csNodeTypeCompany;
          ftUSER        : cObjectType := csNodeTypeUser;
          ftEXT_COMPANY : cObjectType := csNodeTypeExtCompany;
          ftEXT_USER    : cObjectType := csNodeTypeExtUser;
          else
            cObjectType := csNodeTypeDynamic;
        end
      end
    end
    else
      if fInternal.m_type = CCT_RESULT then
        cObjectType := csObjectTypeResultItem;

  if result = S_OK then
    result := CreateData (cObjectType, (lstrlenw (cObjectType) + 1) * sizeof (WideChar), medium)
end;

function TDataObject.CreateDisplayName (var medium : TStgMedium) : HRESULT;
var
  szDispName : WideString;
begin
  szDispName := idsNodeName;
  result := CreateData (PWideChar (szDispName), (Length (szDispName) + 1) * sizeof (WideChar), medium);
end;

function TDataObject.CreateInternal (var medium : TstgMedium) : HRESULT;
begin
  result := CreateData (@fInternal, sizeof (fInternal), medium);
end;

function TDataObject.CreateWorkstationName (var medium : TstgMedium) : HRESULT;
var
  buffer : array [0..MAX_COMPUTERNAME_LENGTH] of char;
  bufLen : DWORD;
  wComputerName : WideString;
begin
  bufLen := sizeof (buffer);
  if GetComputerName (buffer, bufLen) then
  begin
    wComputerName := buffer;
    result := CreateData (PWideChar (wComputerName), (Length (wComputerName) + 1) * sizeof (WideChar), medium);
  end
  else
    result := E_FAIL
end;

function TDataObject.GetData(const formatetcIn: TFormatEtc; out medium: TStgMedium): HResult;
var
  pb : PInteger;
begin
  result := DV_E_CLIPFORMAT;

  if formatetcIn.cfFormat = s_cfMultiSel then
  begin
    Assert (fInternal.m_cookie = MMC_MULTI_SELECT_COOKIE);
    if fInternal.m_cookie <> MMC_MULTI_SELECT_COOKIE then
      result := E_FAIL
    else
    begin
      ASSERT( fpMultiSelData <> Nil);
      ASSERT (fcbMultiSelData <> 0);

      medium.tymed := TYMED_HGLOBAL;
      medium.hGlobal := GlobalAlloc (GMEM_SHARE or GMEM_MOVEABLE, (fcbMultiSelData + sizeof(DWORD)));

      if medium.hGlobal = 0 then
        result := STG_E_MEDIUMFULL
      else
      begin
        pb := GlobalLock (medium.hGlobal);
        pb^ := fcbMultiSelData div sizeof(TGuid);
        Inc (pb);
        Move (fpMultiSelData, pb^, fcbMultiSelData);
        GlobalUnlock (medium.hGlobal);
        result := S_OK
      end
    end
  end
end;

function TDataObject.EnumFormatEtc (dwDirection: Longint; out enumFormatEtc: IEnumFormatEtc): HResult;
begin
  result := E_NOTIMPL
end;

function TDataObject.QueryGetData(const formatetc: TFormatEtc): HResult;
begin
  result := E_NOTIMPL
end;

function TDataObject.GetCanonicalFormatEtc(const formatetc: TFormatEtc; out formatetcOut: TFormatEtc): HResult;
begin
  result := E_NOTIMPL
end;

function TDataObject.SetData(const formatetc: TFormatEtc; var medium: TStgMedium; fRelease: BOOL): HResult;
begin
  result := E_NOTIMPL
end;

function TDataObject.DAdvise(const formatetc: TFormatEtc; advf: Longint; const advSink: IAdviseSink; out dwConnection: Longint): HResult;
begin
  result := E_NOTIMPL
end;

function TDataObject.DUnadvise(dwConnection: Longint): HResult;
begin
  result := E_NOTIMPL
end;

function TDataObject.EnumDAdvise(out enumAdvise: IEnumStatData): HResult;
begin
  result := E_NOTIMPL
end;

procedure TDataObject.SetCookie (value : LongInt);
begin
  fInternal.m_cookie := value
end;

function TDataObject.GetCookie : LongInt;
begin
  result := fInternal.m_cookie
end;

procedure TDataObject.SetString (const value : string);
begin
  fInternal.m_string := value
end;

function TDataObject.GetString : string;
begin
  result := fInternal.m_string
end;

procedure TDataObject.SetCLSID (const value : TGuid);
begin
  fInternal.m_clsid := value
end;

function TDataObject.GetCLSID : TGuid;
begin
  result := fInternal.m_clsid
end;

procedure TDataObject.SetObjectType (value : DATA_OBJECT_TYPES);
begin
  fInternal.m_type := value
end;

function TDataObject.GetObjectType : DATA_OBJECT_TYPES;
begin
  result := fInternal.m_type
end;

procedure TDataObject.SetMultiSelData (pbMultiSelData : PBYTE; cbMultiSelData : Integer);
begin
  fpMultiSelData := pbMultiSelData;
  fcbMultiSelData := cbMultiSelData
end;

procedure TDataObject.SetMultiSelDobj;
begin
  fbMultiSelDobj := TRUE;
end;

function ExtractFromDataObject (dataObject : IDataObject; cf : UINT; cb : LongInt; var hglob : HGLOBAL) : HRESULT;
var
  stgmedium : TStgMedium;
  formatetc : TFormatETC;

begin
  stgmedium.tymed := TYMED_HGLOBAL;
  stgmedium.hBitmap := 0;

  formatetc.cfFormat := cf;
  formatetc.ptd := Nil;
  formatetc.dwAspect := DVASPECT_CONTENT;
  formatetc.lindex := -1;
  formatetc.tymed := TYMED_HGLOBAL;

  Assert (dataObject <> Nil);
  hglob := 0;

  stgmedium.hGlobal := GlobalAlloc (GMEM_SHARE, cb);
  if stgmedium.hGlobal = 0 then
    result := E_OUTOFMEMORY
  else
  begin
    result := dataObject.GetDataHere (formatetc, stgmedium);
    if result = S_OK then
    begin
      hglob := stgmedium.hGlobal;
      stgmedium.hGlobal := 0
    end
  end;

  if (result <> S_OK) and (stgmedium.hGlobal <> 0) then
    GlobalFree (stgmedium.hGlobal);
end;

function ExtractOwnDataObject (dataObject : IDataObject)  : TDataObject;
var
  hglob : HGLOBAL;
  pdo : TDataObject;
  hr : HRESULT;
begin
  pdo := Nil;
  hr := ExtractFromDataObject (dataObject, s_cfInternal, sizeof (PPDataObject), hglob);
  if hr = S_OK then
  begin
    pdo := PDataObject (hGlob)^;
    Assert (Assigned (pdo));
    GlobalFree (hGlob);
  end;
  result := pdo
end;

function ExtractClassId (dataObject : IDataObject) : PGuid;
var
  data : HGLOBAL;
  rc : HRESULT;
begin
  rc := ExtractFromDataObject (dataObject, s_cfCoClass, sizeof (TGuid), data);
  if rc = S_OK then
    result := PGuid (data)
  else
    result := nil
end;

function ExtractNodeType (dataObject : IDataObject) : PGuid;
var
  data : HGLOBAL;
  rc : HRESULT;
begin
  rc := ExtractFromDataObject (dataObject, s_cfNodeType, sizeof (TGuid), data);
  if rc = S_OK then
    result := PGuid (data)
  else
    result := Nil
end;

function ExtractWorkstation (dataObject : IDataObject) : PWideChar;
var
  data : HGLOBAL;
  rc : HRESULT;
  len : Integer;
begin
  len := sizeof (WideChar) * (MAX_COMPUTERNAME_LENGTH + 1);
  rc := ExtractFromDataObject (dataObject, s_cfWorkstation, len, data);
  if rc = S_OK then
    result := PWideChar (data)
  else
    result := Nil
end;

function ExtractInternalFormat (dataObject : IDataObject) : PInternal;
var
  data : HGLOBAL;
  rc : HRESULT;
begin
  rc := ExtractFromDataObject (dataObject, s_cfInternal, sizeof (TInternal), data);
  if rc = S_OK then
    result := PInternal (data)
  else
    result := Nil
end;

end.
