unit MMCFactory;

interface

uses Windows, ComObj, ThreadedFactory;

type
TMMCObjectFactory = class (TThreadedAutoObjectFactory)
private
  fSnapinName : string;
public
  constructor Create(const ASnapinName : string; ComServer: TComServerObject; AutoClass: TAutoClass; const ClassID: TGUID; Instancing: TClassInstancing);
  procedure UpdateRegistry (Register : boolean); override;
  property SnapinName : string read fSnapinName;
end;

implementation

procedure CreateMMCKey(const Key, ValueName, Value: string);
var
  Handle: HKey;
  Status, Disposition: Integer;
  MMCKey : string;
begin
  MMCKey := 'SOFTWARE\Microsoft\MMC';
  if Key [1] <> '\' then
    MMCKey := MMCKey + '\';
  MMCKey := MMCKey + Key;
  Status := RegCreateKeyEx(HKEY_LOCAL_MACHINE, PChar(MMCKey), 0, '',
    REG_OPTION_NON_VOLATILE, KEY_READ or KEY_WRITE, nil, Handle,
    @Disposition);
  if Status = 0 then
  begin
    Status := RegSetValueEx(Handle, PChar(ValueName), 0, REG_SZ,
      PChar(Value), Length(Value) + 1);
    RegCloseKey(Handle);
  end;
  if Status <> 0 then raise EOleError.Create('Unable to create snapin key');
end;

procedure DeleteMMCKey(const Key: string);
var
  MMCKey : string;
begin
  MMCKey := 'SOFTWARE\Microsoft\MMC';
  if Key [1] <> '\' then
    MMCKey := MMCKey + '\';
  MMCKey := MMCKey + Key;
  RegDeleteKey(HKEY_LOCAL_MACHINE, PChar(MMCKey));
end;

constructor TMMCObjectFactory.Create(const ASnapinName : string; ComServer: TComServerObject; AutoClass: TAutoClass; const ClassID: TGUID; Instancing: TClassInstancing);
begin
  fSnapinName := ASnapinName;
  inherited Create (ComServer, AutoClass, ClassID, Instancing, tmApartmentThreading);
end;

procedure TMMCObjectFactory.UpdateRegistry (Register : boolean);
var
  snapinKey : string;
begin
  snapinKey := 'Snapins\' + GUIDToString (ClassID);
  if Register then
  begin
    inherited UpdateRegistry (Register);
    CreateMMCKey (snapinKey, '', SnapinName);
    CreateMMCKey (snapinKey + '\StandAlone', '', '');
    CreateMMCKey (snapinKey, 'Provider', 'Woozle Software');
    CreateMMCKey (snapinKey, 'NameString', SnapinName);
  end
  else
  begin
    DeleteMMCKey (snapinKey + '\StandAlone');
    DeleteMMCKey (snapinKey);
    inherited UpdateRegistry (Register);
  end

end;

end.
