unit TestSnapin;

interface

uses
  windows, classes, ComObj, ActiveX, MMCFactory, TestLib_TLB, mmc, DataObject, Service;

type
  TTTestSnapin = class(TAutoObject, ITTestSnapin, IComponentData, IExtendPropertySheet, IExtendContextMenu, IPersistStream)
  private
    fConsoleNameSpace : IConsoleNameSpace;
    fConsole : IConsole;
    fScopeItemList : TList;
    bmp16x16 : HBITMAP;
    bmp32x32 : HBITMAP;
    m_bIsDirty : boolean;
    m_bInitialized : boolean;

  protected
    function IComponentData.Initialize = ComponentDataInitialize;
    function IComponentData.Destroy = ComponentDataDestroy;
    function IComponentData.GetDisplayInfo = ComponentDataGetDisplayInfo;
    function IComponentData.CompareObjects = ComponentDataCompareObjects;
    function IComponentData.CreateComponent = ComponentDataCreateComponent;
    function IComponentData.Notify = ComponentDataNotify;
    function IComponentData.QueryDataObject = ComponentDataQueryDataObject;

    function IExtendPropertySheet.CreatePropertyPages = ExtendPropertySheetCreatePropertyPages;
    function IExtendPropertySheet.QueryPagesFor = ExtendPropertySheetQueryPagesFor;

    function IExtendContextMenu.Command = ExtendContextMenuCommand;
    function IExtendContextMenu.AddMenuItems = ExtendContextMenuAddMenuItems;

    function IPersistStream.IsDirty = PersistStreamIsDirty;
    function IPersistStream.Load = PersistStreamLoad;
    function IPersistStream.Save = PersistStreamSave;
    function IPersistStream.GetSizeMax = PersistStreamGetSizeMax;
    function IPersistStream.GetClassID = PersistStreamGetClassID;

    function ComponentDataInitialize (pUnknown : IUnknown) : HResult; stdcall;
    function ComponentDataDestroy : HResult; stdcall;
    function ComponentDataGetDisplayInfo (var scopeDataItem : TScopeDataItem) : HResult; stdcall;
    function ComponentDataCompareObjects (const dataObjectA, dataObjectB : IDataObject) : HResult; stdcall;
    function ComponentDataCreateComponent (out component : IComponent) : HResult; stdcall;
    function ComponentDataNotify (const dataObject : IDataObject; event : MMC_NOTIFY_TYPE; arg, param : LongInt) : HResult; stdcall;
    function ComponentDataQueryDataObject(cookie : LongInt; types : DATA_OBJECT_TYPES; out dataObject : IDataObject) : HResult; stdcall;

    function ExtendPropertySheetCreatePropertyPages (provider : IPropertySheetCallback; handle : LongInt; dataObject : IDataObject) : HResult; stdcall;
    function ExtendPropertySheetQueryPagesFor (dataObject : IDataObject) : HResult; stdcall;

    // IPersistStream
    function PersistStreamIsDirty: HResult; stdcall;
    function PersistStreamLoad (const stm: IStream): HResult; stdcall;
    function PersistStreamSave (const stm: IStream; fClearDirty: BOOL): HResult; stdcall;
    function PersistStreamGetSizeMax (out cbSize: Largeint): HResult; stdcall;
    function PersistStreamGetClassID (out classID: TCLSID): HResult; stdcall;

    function OnExpand (dataObject : IDataObject; arg, param : LongInt) : HRESULT;
    function OnProperties (param : LongInt) : HRESULT;
    function OnDelete (cookie : Integer) : HRESULT;
    function OnRemoveChildren (arg : LongInt) : HRESULT;
    function OnRename (cookie, arg, param : LongInt) : HRESULT;

    procedure EnumerateScopePane (dataObject : IDataObject; pParent : HSCOPEITEM);
    procedure CreateFolderList (dataObject : IDataObject);

  public
    fStaticRoot : HSCOPEITEM;
    constructor Create;

    function FindObject (cookie : Integer) : TFolder;
    function GetCoClassID : TGuid; virtual;

    function ExtendContextMenuCommand (nCommandID : Integer; const dataObject : IDataObject) : HRESULT; stdcall;
    function ExtendContextMenuAddMenuItems (const dataObject : IDataObject;  const piCallback: IContextMenuCallback; var pInsertionAllowed: Integer) : HRESULT; stdcall;
  end;

function _QueryDataObject (cookie : LongInt; tp : DATA_OBJECT_TYPES; pImpl : TTTestSnapin; var dataObject : IDataObject) : HRESULT;

implementation

{$R TestLibBitmaps.res}

uses ComServ, comp;

const
  FOLDER_IMAGE_IDX = 0;
  OPEN_FOLDER_IMAGE_IDX = 6;

  FolderData : array [0..NUM_FOLDERS - 1] of TFolderData = (
    (szName : 'User Data'; szSize : '1111'; szType : 'Info about users'; type_ : ftUSER),
    (szName : 'Company Data'; szSize : '2222'; szType : 'Info about Companies'; type_ : ftCOMPANY),
    (szName : 'Virtual Data'; szSize : '3333'; szType : 'Info about virtual items'; type_ : ftVIRTUAL),
    (szName : ''; szSize : ''; szType : ''; type_ : ftSTATIC)
  );

  ExtFolderData : array [0..NUM_FOLDERS - 1] of TFolderData = (
    (szName : '1'; szSize : '1111'; szType : 'Info about users'; type_ : ftEXT_USER),
    (szName : '2'; szSize : '2222'; szType : 'Info about Companies'; type_ : ftEXT_COMPANY),
    (szName : '3'; szSize : '3333'; szType : 'Info about virtual items'; type_ : ftEXT_VIRTUAL),
    (szName : ''; szSize : ''; szType : ''; type_ : ftSTATIC)
  );

constructor TTTestSnapin.Create;
begin
  m_bIsDirty := True;
end;

function TTTestSnapin.ComponentDataInitialize (pUnknown : IUnknown) : HResult;
var
  imageList : IImageList;
begin
  result := S_OK;
  Assert (Assigned (pUnknown));
  Assert (not Assigned (fConsoleNameSpace));
  OleCheck (pUnknown.QueryInterface (IConsoleNameSpace, fConsoleNameSpace));
  OleCheck (pUnknown.QueryInterface (IConsole, fConsole));
  OleCheck (fConsole.QueryScopeImageList (imageList));

  bmp16x16 := LoadBitmap (hInstance, 'BMP16x16');
  bmp32x32 := LoadBitmap (hInstance, 'BMP32x32');

  fScopeItemList := TList.Create;

  Assert (bmp16x16 <> 0);
  Assert (bmp32x32 <> 0);

  m_bInitialized := True;

  OleCheck (imageList.ImageListSetStrip (bmp16x16, bmp32x32, 0, RGB (255, 255, 255)));
end;

function TTTestSnapin.ComponentDataDestroy : HResult;
var
  i : Integer;
begin
  fConsoleNameSpace := Nil;
  fConsole := Nil;
  Assert (DeleteObject(bmp16x16));
  Assert (DeleteObject(bmp32x32));
  for i := 0 to fScopeItemList.Count - 1 do
    TFolder (fScopeItemList [i]).Free;
  fScopeItemList.Free;
  result := S_OK;
end;

function TTTestSnapin.ComponentDataGetDisplayInfo (var scopeDataItem : TScopeDataItem) : HResult;
var
  folder : TFolder;
begin
  folder := TFolder (scopeDataItem.lParam);
  Assert ((scopeDataItem.mask and SDI_STR) <> 0);
  scopeDataItem.displayName := folder.FolderName;
  result := S_OK
end;

function TTTestSnapin.ComponentDataCompareObjects (const dataObjectA, dataObjectB : IDataObject) : HResult;
var
  objA, objB : PInternal;
begin
  result := S_FALSE;

  objA := ExtractInternalFormat (dataObjectA);
  if Assigned (objA) then
  try
    objB := ExtractInternalFormat (dataObjectB);

    if Assigned (objB) then
    try
      if ObjA = ObjB then
        result := S_OK;
    finally
      GlobalFree (THandle (ObjB))
    end
    else Assert (True);
  finally
    GlobalFree (THandle (ObjA))
  end
  else Assert (True);
end;

function TTTestSnapin.ComponentDataCreateComponent (out component : IComponent) : HResult;
var
  comp : TIComponent;
begin
  comp :=TIComponent.Create;
  comp.SetIComponentData (self);
  component := comp as IComponent;
end;

function TTTestSnapin.ComponentDataNotify (const dataObject : IDataObject; event : MMC_NOTIFY_TYPE; arg, param : LongInt) : HResult;
var
  Internal : PInternal;
  cookie : LongInt;

begin
  result := S_OK;
  Assert (fConsoleNameSpace <> Nil);

    // Since it's my folder it has an internal format.
    // Design Note: for extension.  I can use the fact, that the data object doesn't have
    // my internal format and I should look at the node type and see how to extend it.
  if event = MMCN_PROPERTY_CHANGE then
    result := OnProperties (param)
  else
  begin
    Internal := ExtractInternalFormat (dataObject);
    if Assigned (Internal) then
    try
      cookie := Internal^.m_cookie;

      case event of
        MMCN_PASTE           : MessageBox (0, 'TTTestSnapin.MMCN_PASTE', Nil, MB_ICONINFORMATION);
        MMCN_DELETE          : Result := OnDelete (cookie);
        MMCN_REMOVE_CHILDREN : Result := OnRemoveChildren (arg);
        MMCN_RENAME          : Result := OnRename (cookie, arg, param);
        MMCN_EXPAND          : Result := OnExpand (dataObject, arg, param)
      end
    finally
      GlobalFree (THandle (Internal))
    end
  end
end;

function _QueryDataObject (cookie : LongInt; tp : DATA_OBJECT_TYPES; pImpl : TTTestSnapin; var dataObject : IDataObject) : HRESULT;
var
  pObject : TDataObject;
begin
  Assert (Assigned (pImpl));
  pObject := TDataObject.Create;
  pObject.ObjectType := tp;
  pObject.Cookie := cookie;
  pObject.CLSID := pImpl.GetCoClassID;
  dataObject := pObject as IDataObject;
  result := S_OK;
end;

function TTTestSnapin.ComponentDataQueryDataObject(cookie : LongInt; types : DATA_OBJECT_TYPES; out dataObject : IDataObject) : HResult;
begin
  result := _QueryDataObject (cookie, types, self, dataObject);
end;

function TTTestSnapin.GetCoClassID : TGuid;
begin
  result := Class_TTestSnapin
end;

function TTTestSnapin.OnExpand (dataObject : IDataObject; arg, param : LongInt) : HRESULT;
begin
  if BOOL (arg) = TRUE then
  begin
    Assert (Assigned (fConsoleNameSpace));
    EnumerateScopePane (dataObject, param)
  end;

  result := S_OK;
end;

function TTTestSnapin.OnProperties (param : LongInt) : HRESULT;
begin
  MessageBox (0, 'OnProperties', Nil, MB_ICONINFORMATION);
  result := S_OK;
end;

function TTTestSnapin.OnDelete (cookie : Integer) : HRESULT;
begin
  MessageBox (0, 'OnDelete', Nil, MB_ICONINFORMATION);
  result := S_OK;
end;

function TTTestSnapin.OnRemoveChildren (arg : LongInt) : HRESULT;
begin
//  MessageBox (0, 'OnRemoveChildren', Nil, MB_ICONINFORMATION);
  result := S_OK;
end;

function TTTestSnapin.OnRename (cookie, arg, param : LongInt) : HRESULT;
var
  pszNewName : PWideChar;
  folder : TFolder;
begin
  result := S_OK;
  if arg = 0 then
  begin
    pszNewName := PWideChar (param);
    if not Assigned (pszNewName) then
      result := E_INVALIDARG
    else
    begin
      folder := TFolder (cookie);
      Assert (Assigned (folder));
      if not Assigned (folder) then
        result := E_INVALIDARG
      else
        folder.FolderName := pszNewName
    end
  end
end;

procedure TTTestSnapin.CreateFolderList (dataObject : IDataObject);
var
  folder : TFolder;
  workstation : PWideChar;
  coClassID : PGuid;
  folderSet : PFolderData;
  bExtend : BOOL;
  i : Integer;

begin
  Assert (Assigned (dataObject));

  workstation := ExtractWorkstation (dataObject);
  Assert (Assigned (workstation));

  coClassId := ExtractClassId (dataObject);
  Assert (Assigned (coClassID));

  bExtend := not IsEqualCLSID (coClassId^, GetCoClassID);

  if bExtend then
    folderSet := @ExtFolderData [0]
  else
    folderSet := @FolderData [0];

  Assert (fScopeItemList.Count = 0);

  for i := 0 to NUM_FOLDERS - 1 do
  begin
    folder := TFolder.Create;
    folder.Init (folderSet^.szName, FOLDER_IMAGE_IDX, OPEN_FOLDER_IMAGE_IDX, folderSet^.type_, False);
    Inc (folderSet);
    fScopeItemList.Add (folder)
  end;
  folder.Cookie := 0;

  GlobalFree (THandle (workstation));
  GlobalFree (THandle (coClassID))
end;

function TTTestSnapin.FindObject (cookie : Integer) : TFolder;
var
  i : Integer;
begin
  result := Nil;
  for i := 0 to fScopeItemList.Count - 1 do
    if TFolder (fScopeItemList [i]).cookie = cookie then
    begin
      result := TFolder (fScopeItemList [i]);
      break
    end
end;

procedure TTTestSnapin.EnumerateScopePane (dataObject : IDataObject; pParent : HSCOPEITEM);
var
  i : Integer;
  internal : PInternal;
  cookie : LongInt;
  static : TFolder;
  folder : TFolder;
begin
  Assert (Assigned (dataObject));
  Assert (Assigned (dataObject));

  Internal := ExtractInternalFormat (dataObject);

  if Assigned (internal) then
  try
    cookie := internal^.m_cookie;

    if cookie = 0 then
    begin

    // Initialize folder list if empty

      if fScopeItemList.Count = 0 then
        CreateFolderList(dataObject);

    // Enumerate the scope pane
    // return the folder object that represents the cookie
    // Note - for large list, use dictionary

      static := FindObject (cookie);

      Assert (not static.Enumerated);
      Assert (pParent <> 0);

    // Cache the HSCOPEITEM of the static root.
      if cookie = 0 then
        fStaticRoot := pParent;

      for i := 0 to NUM_FOLDERS - 2 do
      begin
        if i >= fScopeItemList.Count then
          Assert (True);

        folder :=  TFolder (fScopeItemList [i]);
        folder.ScopeItem^.relativeID := pParent;
        folder.ScopeItem^.mask := folder.ScopeItem^.mask or SDI_PARAM;
        folder.ScopeItem^.lParam := LongInt (folder);
        folder.Cookie := LongInt (folder);
        fConsoleNameSpace.InsertItem (folder.ScopeItem^);
        Assert (folder.ScopeItem^.ID <> 0)
      end;

      static.Enumerated := True;
      static.ScopeItem^.relativeID := pParent
    end
  finally
    GlobalFree (THandle (Internal))
  end
end;

function TTTestSnapin.ExtendContextMenuCommand (nCommandID : Integer; const dataObject : IDataObject) : HRESULT;
var
  internal : PInternal;
  folder : TFolder;
  retVal : Integer;
begin
    // Note - snap-ins need to look at the data object and determine
    // in what context the command is being called.

    // Handle each of the commands.
  case nCommandID of
    IDM_COMMAND1 :
      begin
        internal := ExtractInternalFormat (dataObject);
        if Assigned (internal) then
        try
          folder := TFolder (internal.m_cookie);
          if Assigned (folder) then
            fConsole.SelectScopeItem (folder.ScopeItem^.ID)
        finally
          GlobalFree (THandle (internal))
        end
        else Assert (False);
      end;
    IDM_COMMAND2:
      begin
        Assert (Assigned (fConsole));
        fConsole.MessageBox ('Snapin Menu Comand Selected',  menuItems [nCommandID].strName, MB_OK, retVal)
      end;

    else
      Assert (False)
  end;
  result := S_OK;
end;

function TTTestSnapin.ExtendContextMenuAddMenuItems (const dataObject : IDataObject;  const piCallback: IContextMenuCallback; var pInsertionAllowed: Integer) : HRESULT;
var
  internal : PInternal;
  bCmd1IsDefault : boolean;
  pMenuItem : PContextMenuItem;

begin
  if IsMMCMultiSelectDataObject(dataObject) then
    result := S_FALSE
  else
  begin
    result := S_OK;
    internal := ExtractInternalFormat (dataObject);
    if Assigned (internal) then
    try
      bCmd1IsDefault := internal^.m_type = CCT_RESULT;

      if bCmd1IsDefault then
        menuItems [0].fSpecialFlags := CCM_SPECIAL_DEFAULT_ITEM
      else
        menuItems [0].fSpecialFlags := 0;

    // Loop through and add each of the menu items

      if (pInsertionAllowed and CCM_INSERTIONALLOWED_TOP) <> 0 then
      begin
        pMenuItem := @menuItems [0];
        while pMenuItem^.strName <> Nil do
        begin
          result := piCallback.AddItem (pMenuItem^);
          if result <> S_OK then
            break
          else
            Inc (pMenuItem)
        end
      end;

    // Loop through and add each of the view items
      if (pInsertionAllowed and CCM_INSERTIONALLOWED_VIEW) <> 0 then
      begin
        pMenuItem := @viewItems [0];
        while pMenuItem^.strName <> Nil do
        begin
          result := piCallback.AddItem (pMenuItem^);
          if result <> S_OK then
            break
          else
            Inc (pMenuItem)
        end
      end
    finally
      GlobalFree (THandle (internal));
    end
  end
end;

function TTTestSnapin.ExtendPropertySheetCreatePropertyPages (provider : IPropertySheetCallback; handle : LongInt; dataObject : IDataObject) : HResult;
begin
  result := S_FALSE;
end;

function TTTestSnapin.ExtendPropertySheetQueryPagesFor (dataObject : IDataObject) : HResult;
begin
  result := S_FALSE;
end;

function TTTestSnapin.PersistStreamIsDirty: HResult;
begin
  if m_bIsDirty then
    result := S_OK
  else
    result := S_FALSE;
end;

function TTTestSnapin.PersistStreamLoad (const stm: IStream): HResult;
var
  psz : array [0..9] of char;
  nBytesRead : Longint;
begin
  Assert (m_bInitialized);

  result := stm.Read (@psz [0], 10, @nBytesRead);
  Assert (Succeeded (result) and (nBytesRead = 10));
  Assert (lstrcmp (psz, '987654321') = 0);
  m_bIsDirty := False;
  if Succeeded (result) then
    result := S_OK
  else
    result := E_FAIL
end;

function TTTestSnapin.PersistStreamSave (const stm: IStream; fClearDirty: BOOL): HResult;
var
  nBytesWritten : Longint;
begin
  Assert (m_bInitialized);
  result := stm.Write (PChar ('987654321'), 10, @nBytesWritten);

  Assert (Succeeded (result) and (nBytesWritten = 10));
  if Failed (result) then
    result := STG_E_CANTSAVE
  else
  begin
    m_bIsDirty := False;
    result := S_OK
  end
end;

function TTTestSnapin.PersistStreamGetSizeMax (out cbSize: Largeint): HResult;
begin
  cbSize := 10;
end;

function TTTestSnapin.PersistStreamGetClassID (out classID: TCLSID): HResult; stdcall;
begin
  classID := Class_TTestSnapin;

  result := E_NOTIMPL;
end;



initialization
  TMMCObjectFactory.Create('Test Snapin', ComServer, TTTestSnapin, Class_TTestSnapin, ciMultiInstance);
end.
