unit ThreadedFactory;

interface

uses comobj;

type
TThreadingModel = (tmSingleThreading, tmApartmentThreading, tmFreeThreading, tmBothThreading);

TThreadedAutoObjectFactory = class (TAutoObjectFactory)
private
  fThreadingModel : TThreadingModel;

public
  constructor Create(ComServer: TComServerObject; AutoClass: TAutoClass; const ClassID: TGUID; Instancing: TClassInstancing; Threading : TThreadingModel);
  procedure UpdateRegistry (Register : boolean); override;
  property ThreadingModel : TThreadingModel read fThreadingModel;
end;

implementation

constructor TThreadedAutoObjectFactory.Create(ComServer: TComServerObject; AutoClass: TAutoClass; const ClassID: TGUID; Instancing: TClassInstancing; Threading : TThreadingModel);
begin
  fThreadingModel := Threading;
  inherited Create (ComServer, AutoClass, ClassID, Instancing);
end;

procedure TThreadedAutoObjectFactory.UpdateRegistry (Register : boolean);
var
  classKey : string;
  modelName : string;
begin
  ClassKey := 'CLSID\' + GUIDToString(ClassID);
  if Register then
  begin
    inherited UpdateRegistry (Register);
    case fThreadingModel of
      tmSingleThreading : modelName := '';
      tmApartmentThreading : modelName := 'Apartment';
      tmFreeThreading : modelName := 'Free';
      tmBothThreading : modelName := 'Both';
    end;
    if modelName <> '' then
      CreateRegKey (classKey + '\InprocServer32', 'ThreadingModel', modelName);
  end
  else
  begin
    DeleteRegKey (classKey + '\InprocServer32\ThreadingModel');
    inherited UpdateRegistry (Register);
  end
end;

end.
