unit service;

interface
uses Windows, ActiveX, ComObj, mmc;

const
  UNINITIALIZED = -1;
  NUM_FOLDERS = 4;
  NUM_NAMES = 4;
  NUM_COMPANY = 6;
  NUM_VIRTUAL_ITEMS = 100000;
  MAX_ITEM_NAME = 64;
  SCOPE_ITEM =    111;
  RESULT_ITEM =   222;


  // FOLDER_TYPES enum
  ftSTATIC = $8000;
  ftCOMPANY = $8001;
  ftUSER = $8002;
  ftVIRTUAL = $8003;
  ftEXT_COMPANY = $8004;
  ftEXT_USER = $8005;
  ftEXT_VIRTUAL = $8006;
  ftNONE = $FFFF;

type
  FOLDER_TYPES = Integer;

  TFolderData = record
    szName : PWideChar;
    szSize : PWideChar;
    szType : PWideChar;
    type_ : FOLDER_TYPES
  end;
  PFolderData = ^TFolderData;

  TResultData = record
    parentType : FOLDER_TYPES;
    szName : PWideChar;
    szSize : PWideChar;
    szType : PWideChar
  end;
  PResultData = ^TResultData;

TFolder = class
private
   // Used for debug purpose only. This should be the first
   // member. The class should not have any virtual fuctions.
//  m_ItemType : DWORD;

  m_pScopeItem : PScopeDataItem;
  m_cookie : LongInt;
  m_enumed : BOOL;
  m_type : FOLDER_TYPES;
  m_pszName : PWideChar;

  procedure SetFolderName (value : PWideChar);


public
  constructor Create;
  destructor Destroy; override;
  procedure Init (szName : PWideChar; nImage, nOpenImage : Integer; type_ : FOLDER_TYPES; bHasChildren : BOOL);
//  property ItemType : Integer read m_ItemType;
  property Enumerated : BOOL read m_enumed write m_enumed;
  property Cookie : LongInt read m_cookie write m_cookie;
  property FolderType : FOLDER_TYPES read m_type;
  property ScopeItem : PScopeDataItem read m_pScopeItem;

  function Equals (rhs : TFolder) : boolean;
  property FolderName : PWideChar read m_pszName write SetFolderName;
end;

implementation

constructor TFolder.Create;
begin
//  m_ItemType := SCOPE_ITEM;  // used for debug purpose only

  m_cookie := UNINITIALIZED;
  m_enumed := FALSE;
  m_pScopeItem := Nil;
  m_type := ftNONE;
  m_pszName := Nil
end;

destructor TFolder.Destroy;
begin
  FolderName := Nil;
  inherited
end;

procedure TFolder.Init (szName : PWideChar; nImage, nOpenImage : Integer; type_ : FOLDER_TYPES; bHasChildren : BOOL);
begin
  Assert (not Assigned (m_pScopeItem));
  m_pScopeItem := new (PScopeDataItem);
  FillChar (m_pScopeItem^, Sizeof (m_pScopeItem^), 0);
  m_type := type_;

  if Assigned (szName) then
  begin
    m_pScopeItem^.mask := SDI_STR;
    m_pScopeItem^.displayname := MMC_CALLBACK;
    FolderName := szName
  end;

  if nImage <> 0 then
  begin
    m_pScopeItem^.mask := m_pScopeItem^.mask or SDI_IMAGE;
    m_pScopeItem^.nImage := nImage
  end;

  if nOpenImage <> 0 then
  begin
    m_pScopeItem^.mask := m_pScopeItem^.mask or SDI_OPENIMAGE;
    m_pScopeItem^.nOpenImage := nOpenImage
  end;

  if bHasChildren then
  begin
    m_pScopeItem^.mask := m_pScopeItem^.mask or SDI_CHILDREN;
    m_pScopeItem^.cChildren := 1
  end
end;

function TFolder.Equals (rhs : TFolder) : boolean;
begin
  result := cookie = rhs.cookie
end;

procedure TFolder.SetFolderName (value : PWideChar);
begin
  if Assigned (m_pszName) then
  begin
    CoTaskMemFree (m_pszName);
    m_pszName := Nil
  end;

  if Assigned (value) then
    m_pszName := StringToLPOLESTR (value);
end;

end.
