unit unitCodeSnippets;

interface

uses Windows, Classes, SysUtils, ComObj, ActiveX, ConTnrs;

type
TCodeSnippet = class
private
  fSnippetName: string;
  fCode: string;
public
  constructor Create (const ASnippetName, ACode : string);
  property SnippetName : string read fSnippetName;
  property Code : string read fCode;
end;

TCodeSnipets = class (TObjectList)
private
  procedure LoadSnipets (instance : THandle);
public
  constructor Create (AModuleInstance : THandle);
  function GetSnippet (const snippetName : string) : string;
  function FormatSnippet (const s : string; const ss : array of string; const rs : array of string) : string;
end;

implementation

{ TCodeSnippet }

constructor TCodeSnippet.Create(const ASnippetName, ACode: string);
begin
  fSnippetName := ASnippetName;
  fCode := ACode;
end;

{ TCodeSnipets }

constructor TCodeSnipets.Create(AModuleInstance: THandle);
begin
  inherited Create;
  LoadSnipets (AModuleInstance);
end;

function TCodeSnipets.FormatSnippet(const s: string; const ss,
  rs: array of string): string;
var
  tempStr : string;
  idx : Integer;
  max : Integer;

  function CreateGUID : string;
  var
    r : TGUID;
  begin
    CoCreateGUID (r);
    result := GUIDToString (r);
  end;

begin
  max := High (ss);
  if High (rs) < max then
    max := High (rs);

  result := s;
  for idx := 0 to max do
    result := StringReplace (result, ss [idx], rs [idx], [rfReplaceAll, rfIgnoreCase]);

  repeat
    tempStr := StringReplace (result, '%CreateGUID%', CreateGuid, [rfIgnoreCase]);
    if tempStr = result then
      break;
    result := tempStr
  until False;
end;

function TCodeSnipets.GetSnippet(const snippetName: string): string;
var
  i : Integer;
begin
  result := '';
  for i := 0 to Count - 1 do
    if CompareText (snippetName, TCodeSnippet (Items [i]).SnippetName) = 0 then
    begin
      result := TCodeSnippet (Items [i]).Code;
      break
    end;

  if result = '' then
    raise Exception.Create ('Code Snippet ' + snippetName + ' not found')
end;

procedure TCodeSnipets.LoadSnipets(instance: THandle);
var
  resInstance : THandle;
  res : THandle;
  codeResource : THandle;
  size, i, ps : Integer;
  l : TStringList;
  p, buffer : PChar;
  snippet : TStringList;
  snippetName : string;
  s : string;

begin
  ResInstance := FindResourceHInstance(instance);
  res := FindResource(ResInstance, 'SNIPETS', RT_RCDATA);
  if res > 0 then
  begin
    size := SizeOfResource (resInstance, res);
    codeResource := LoadResource (ResInstance, res);
    p := LockResource (codeResource);
    buffer := StrAlloc (size + 1);
    system.Move (p^, buffer^, size);
    l := TStringList.Create;
    try
      l.Text := buffer;
      snippet := Nil;
      for i := 0 to l.Count - 1 do
      begin
        s := l.Strings [i];
        if Copy (s, 1, 1) = '-' then
          if Assigned (snippet) then
          begin
            Add (TCodeSnippet.Create (snippetName, snippet.Text));
            FreeAndNil (snippet)
          end
          else
          begin
            ps := Pos (':', s);
            if ps > 0 then
            begin
              snippet := TStringList.Create;
              snippetName := Trim (Copy (s, ps + 1, MaxInt));
            end
          end
        else
          if Assigned (snippet) then
            snippet.Add (s)
      end;

      if Assigned (snippet) then
      begin
        Add (TCodeSnippet.Create (snippetName, snippet.Text));
        FreeAndNil (snippet)
      end
    finally
      l.Free
    end
  end
end;

end.
